{% extends "admin/change_form.html" %}

{% if save_on_top %}
  {% block submit_buttons_top %}
    {% include 'admin/submit_line.html' with show_save=1 %}
  {% endblock %}
{% endif %}

{% block submit_buttons_bottom %}
  {% include 'admin/submit_line.html' with show_save=1 %}
{% endblock %}
